/*
 * (C)opyright 1993,1994,1995 by Darren Reed.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this notice is preserved and due credit is given
 * to the original author and the contributors.
 */
#pragma ident   "@(#)solaris.c	1.9 4/10/96 (C) 1995 Darren Reed"

#include <sys/systm.h>
#include <sys/types.h>
#include <sys/param.h>
#include <sys/errno.h>
#include <sys/uio.h>
#include <sys/buf.h>
#include <sys/modctl.h>
#include <sys/open.h>
#include <sys/kmem.h>
#include <sys/poll.h>
#include <sys/conf.h>
#include <sys/cmn_err.h>
#include <sys/stat.h>
#include <sys/cred.h>
#include <sys/dditypes.h>
#include <sys/stream.h>
#include <sys/autoconf.h>
#include <sys/byteorder.h>
#include <sys/socket.h>
#include <net/if.h>
#include <net/af.h>
#include <net/route.h>
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/if_ether.h>
#include <netinet/ip.h>
#include <netinet/ip_var.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>
#include <netinet/tcpip.h>
#include <netinet/ip_icmp.h>
#include "ipl.h"
#include "ip_fil.h"
#include "ip_compat.h"
#include <sys/ddi.h>
#include <sys/sunddi.h>

char	_depends_on[] = "drv/ip";

extern	int	iplopen(), iplclose(), iplread(), iplioctl();
extern	int	iplattach(), ipldetach();

int	solattach(), soldetach();

extern	struct	filterstats	frstats[];
extern	kmutex_t	ipl_mutex, ipf_mutex, ipfs_mutex;

static	qif_t	*qif_head = NULL;
static	int	(*ip_r_putq)() = NULL;
static	int	ip_r_putqs = 0;

static	int	ipl_getinfo(), ipl_probe(), ipl_identify(), ipl_attach();
static	int	ipl_detach();

static struct cb_ops ipl_cb_ops = {
	iplopen,
	iplclose,
	nodev,		/* strategy */ nodev,		/* print */
	nodev,		/* dump */
	iplread,
	nodev,		/* write */
	iplioctl,	/* ioctl */
	nodev,		/* devmap */
	nodev,		/* mmap */
	nodev,		/* segmap */
	nochpoll,	/* poll */
	ddi_prop_op,
	NULL,
	D_MTSAFE
};

static struct dev_ops ipl_ops = {
	DEVO_REV,
	0,
	ipl_getinfo,
	ipl_identify,
	ipl_probe,
	ipl_attach,
	ipl_detach,
	nodev,		/* reset */
	&ipl_cb_ops,
	(struct bus_ops *)0
};

extern struct mod_ops mod_driverops;
static struct modldrv iplmod = {
	&mod_driverops, IPL_VERSION, &ipl_ops };
static struct modlinkage modlink1 = { MODREV_1, &iplmod, 0 };


static dev_info_t *ipf_dev_info = NULL;


int _init()
{
#ifdef	IPFDEBUG
	cmn_err(CE_NOTE, "ipf_init()");
#endif
	return mod_install(&modlink1);
}


int _fini(void)
{
#ifdef	IPFDEBUG
	cmn_err(CE_NOTE, "ipf_fini()");
#endif
	return mod_remove(&modlink1);
}


int _info(struct modinfo *modinfop)
{
#ifdef	IPFDEBUG
	cmn_err(CE_NOTE, "ipf_info(%x)", modinfop);
#endif
	return mod_info(&modlink1, modinfop);
}


static int ipl_probe(dev_info_t *dip)
{
#ifdef	IPFDEBUG
	cmn_err(CE_NOTE, "ipl_probe(%x)", dip);
#endif
	return DDI_PROBE_SUCCESS;
}


static int ipl_identify(dev_info_t *dip)
{
#ifdef	IPFDEBUG
	cmn_err(CE_NOTE, "ipl_identify(%x)", dip);
#endif
	if (strcmp(ddi_get_name(dip), "ipf") == 0)
		return (DDI_IDENTIFIED);
	return (DDI_NOT_IDENTIFIED);
}


static int ipl_attach(dev_info_t *dip, ddi_attach_cmd_t cmd)
{
	int instance;

#ifdef	IPFDEBUG
	cmn_err(CE_NOTE, "ipl_attach(%x,%x)", dip, cmd);
#endif
	switch (cmd) {
	case DDI_ATTACH:
		instance = ddi_get_instance(dip);
#ifdef	IPFDEBUG
		cmn_err(CE_NOTE, "attach ipf instace %d", instance);
#endif
		if (ddi_create_minor_node(dip, "ipf", S_IFCHR, instance,
					  DDI_PSEUDO, 0) == DDI_FAILURE) {
			ddi_remove_minor_node(dip, NULL);
			goto attach_failed;
		}
		ipf_dev_info = dip;
		sync();
		/*
		 * Initialize mutex's
		 */
		iplattach();
		solattach();
		cmn_err(CE_CONT, "IP Filter attaching complete.\n");
		return (DDI_SUCCESS);
	default:
		return (DDI_FAILURE);
	}

attach_failed:
	/*
	 * Use our own detach routine to toss
	 * away any stuff we allocated above.
	 */
	(void) ipl_detach(dip, DDI_DETACH);
	return (DDI_FAILURE);
}


static int ipl_detach(dip, cmd)
dev_info_t *dip;
ddi_detach_cmd_t cmd;
{
	int instance;

#ifdef	IPFDEBUG
	cmn_err(CE_NOTE, "ipl_detach(%x,%x)", dip, cmd);
#endif
	switch (cmd) {
	case DDI_DETACH:
		/*
		 * Undo what we did in ipl_attach, freeing resources
		 * and removing things we installed.  The system
		 * framework guarantees we are not active with this devinfo
		 * node in any other entry points at this time.
		 */
		ddi_prop_remove_all(dip);
		instance = ddi_get_instance(dip);
		ddi_remove_minor_node(dip, NULL);
		if (!soldetach()) {
			cmn_err(CE_CONT, "IP Filter detached\n");
			return (DDI_SUCCESS);
		}
	default:
		return (DDI_FAILURE);
	}
}


static int ipl_getinfo(dip, infocmd, arg, result)
dev_info_t *dip;
ddi_info_cmd_t infocmd;
void *arg, **result;
{
	int error = DDI_FAILURE;

#ifdef	IPFDEBUG
	cmn_err(CE_NOTE, "ipl_getinfo(%x,%x)", dip, infocmd);
#endif
	switch (infocmd) {
	case DDI_INFO_DEVT2DEVINFO:
		*result = ipf_dev_info;
		error = DDI_SUCCESS;
		break;
	case DDI_INFO_DEVT2INSTANCE:
		*result = (void *)getminor((dev_t) arg);
		error = DDI_SUCCESS;
		break;
	default:
		break;
	}
	return (error);
}


/*
 * find the filter structure setup for this queue
 */
qif_t *qif_from_queue(q)
queue_t *q;
{
	qif_t *qif;

	for (qif = qif_head; qif; qif = qif->qf_next)
		if ((qif->qf_iptr == q->q_ptr) ||
		    (qif->qf_optr == q->q_ptr))
			break;
	return qif;
}


/*
 * find the first data mblk, if present, in the chain we're processing.  Also
 * make a few sanity checks to try prevent the filter from causing a panic -
 * none of the nice IP sanity checks (including checksumming) should have been
 * done yet - dangerous!
 */
int fr_precheck(mb, q, qif, out)
mblk_t *mb;
queue_t	*q;
qif_t *qif;
int out;
{
	u_long lbuf[64];
	ip_t *ip;
	int hlen, len, err;
	u_char *s;
	mblk_t *m;

	/*
	 * This at least gets the right stuff for PPP
	 */
	for (m = mb; m && (m->b_datap->db_type != M_DATA); m = mb->b_cont)
		;
	if (!m)
		m = mb;
	s = m->b_rptr;
	len = m->b_wptr - s;

	if (out) {
		s += qif->qf_hl;
		len -= qif->qf_hl;
	}
	bcopy((char *)s, (char *)lbuf, MIN(len, sizeof(lbuf)));
	ip = (ip_t *)lbuf;
#ifdef	IPFDEBUG
/*
	cmn_err(CE_NOTE, "out %d len %d ipv %d ifp %x info %x %x %x %x",
		out, len, ip->ip_v, qif->qf_ill, q->q_qinfo, q, q->q_next,
		q->q_link);
*/
#endif
	if (ip->ip_v == IPVERSION) {
		hlen = ip->ip_hl << 2;
		if ((hlen < sizeof(ip_t)) || (hlen > ntohs(ip->ip_len)) ||
		    (hlen > len)) {
			/*
			 * Bad IP packet - don't filter it but let the IP
			 * stack catch it for statistics.
			 */
			frstats[out].fr_bad++;
			return 0;
		}
#ifndef	sparc
		ip->ip_len = ntohs(ip->ip_len);
		ip->ip_off = ntohs(ip->ip_off);
#endif
		err = fr_check(ip, hlen, qif->qf_ill, out, qif, q);
#ifndef	sparc
		ip->ip_len = htons(ip->ip_len);
		ip->ip_off = htons(ip->ip_off);
#endif
		return err;
	}
	return 0;
}


int fr_qin(q, mb)
queue_t *q;
mblk_t *mb;
{
	int (*pnext)();
	qif_t qfb, *qif;

	mutex_enter(&ipfs_mutex);
	while (!(qif = qif_from_queue(q))) {
		for (qif = qif_head; qif; qif = qif->qf_next)
			if (qif->qf_rqinfo == q->q_qinfo && qif->qf_inp) {
				pnext = qif->qf_inp;
				mutex_exit(&ipfs_mutex);
				frstats[0].fr_notip++;
				return (*pnext)(q, mb);
			}
		mutex_exit(&ipfs_mutex);
		ipfsync();
		cmn_err(CE_WARN, "dropped: fr_qin(%x,%x): type %x qif %x",
			q, mb, mb->b_datap->db_type, qif);
		cmn_err(CE_CONT, "info %x next %x ptr %x fsrv %x bsrv %x\n",
			q->q_qinfo, q->q_next, q->q_ptr, q->q_nfsrv,
			q->q_nbsrv);
		cmn_err(CE_CONT, "info: putp %x srvp %x info %x\n",
			q->q_qinfo->qi_putp, q->q_qinfo->qi_srvp,
			q->q_qinfo->qi_infop);
		frstats[0].fr_drop++;
		freemsg(mb);
		return 0;
	}
	/*
	 * So we can be more re-entrant.
	 */
	pnext = qif->qf_inp;
	bcopy((char *)qif, (char *)&qfb, sizeof(*qif));
	qif = &qfb;
	mutex_exit(&ipfs_mutex);

	if (mb->b_datap->db_type == M_DATA || mb->b_datap->db_type == M_PROTO)
		if (fr_precheck(mb, q, qif, 0)) {
			freemsg(mb);
			return 0;
		}

	if (pnext)
		return (*pnext)(q, mb);

	cmn_err(CE_WARN, "inp NULL: qif %x %s q %x info %x",
		qif, qif->qf_name, q, q->q_qinfo);
	freemsg(mb);
	return 0;
}


int fr_qout(q, mb)
queue_t *q;
mblk_t *mb;
{
	int (*pnext)();
	qif_t qfb, *qif;

	mutex_enter(&ipfs_mutex);
	if (!(qif = qif_from_queue(q))) {
		for (qif = qif_head; qif; qif = qif->qf_next)
			if (qif->qf_wqinfo == q->q_qinfo && qif->qf_outp) {
				pnext = qif->qf_outp;
				mutex_exit(&ipfs_mutex);
				frstats[1].fr_notip++;
				return (*pnext)(q, mb);
			}
		mutex_exit(&ipfs_mutex);
		ipfsync();
		cmn_err(CE_WARN, "dropped: fr_qout(%x,%x): type %x: link %x",
			q, mb, mb->b_datap->db_type, qif, q->q_link);
		cmn_err(CE_CONT, "info %x next %x ptr %x fsrv %x bsrv %x\n",
			q->q_qinfo, q->q_next, q->q_ptr, q->q_nfsrv,
			q->q_nbsrv);
		cmn_err(CE_CONT, "info: putp %x srvp %x info %x\n",
			q->q_qinfo->qi_putp, q->q_qinfo->qi_srvp,
			q->q_qinfo->qi_infop);
		if (q->q_nfsrv)
			cmn_err(CE_CONT, "nfsrv: info %x next %x ptr %x\n",
				q->q_nfsrv->q_qinfo, q->q_nfsrv->q_next,
				q->q_nfsrv->q_ptr);
		if (q->q_nbsrv)
			cmn_err(CE_CONT, "nbsrv: info %x next %x ptr %x\n",
				q->q_nbsrv->q_qinfo, q->q_nbsrv->q_next,
				q->q_nbsrv->q_ptr);
		frstats[1].fr_drop++;
		freemsg(mb);
		return 0;
	}
	/*
	 * So we can be more re-entrant.
	 */
	pnext = qif->qf_outp;
	bcopy((char *)qif, (char *)&qfb, sizeof(*qif));
	qif = &qfb;
	mutex_exit(&ipfs_mutex);

	if (mb->b_datap->db_type == M_DATA || mb->b_datap->db_type == M_PROTO)
		if (fr_precheck(mb, q, qif, 1)) {
			freemsg(mb);
			return 0;
		}

	if (pnext)
		return (*pnext)(q, mb);

	cmn_err(CE_WARN, "outp NULL: qif %x %s q %x info %x",
		qif, qif->qf_name, q, q->q_qinfo);
	freemsg(mb);
	return 0;
}


/*
 * attach the packet filter to each interface that is defined as having an
 * IP address associated with it and save some of the info. for that struct
 * so we're not out of date as soon as te ill disappears - but we must sync
 * to be correct!
 */
int solattach()
{
	queue_t *in, *out;
	qif_t *qif, *qf2;
	ill_t *il;

	for (il = ill_g_head; il; il = il->ill_next) {
		in = il->ill_rq;
		if (!in || !il->ill_wq)
			continue;

		out = il->ill_wq->q_next;

		mutex_enter(&ipfs_mutex);
		/*
		 * Look for entry already setup for this device
		 */
		for (qif = qif_head; qif; qif = qif->qf_next)
			if (qif->qf_iptr == in->q_ptr &&
			    qif->qf_optr == out->q_ptr)
				break;
		if (qif) {
			mutex_exit(&ipfs_mutex);
			continue;
		}
#ifdef	IPFDEBUG
		cmn_err(CE_NOTE, "il %x iptr %x optr %x iput %x oput %x %x/%x",
			il, in->q_ptr,  out->q_ptr, in->q_qinfo->qi_putp,
			out->q_qinfo->qi_putp, out->q_qinfo, in->q_qinfo);
#endif
		qif = (qif_t *)KMALLOC(sizeof(*qif));

		if (in->q_qinfo->qi_putp == fr_qin) {
			for (qf2 = qif_head; qf2; qf2 = qf2->qf_next)
				if (qf2->qf_rqinfo == in->q_qinfo) {
					qif->qf_inp = qf2->qf_inp;
					break;
				}
			if (!qf2) {
#ifdef	IPFDEBUG
				cmn_err(CE_WARN, "rq:%s put %x qi %x",
					il->ill_name, in->q_qinfo->qi_putp,
					in->q_qinfo);
#endif
				mutex_exit(&ipfs_mutex);
				KFREE(qif);
				continue;
			}
		} else
			qif->qf_inp = in->q_qinfo->qi_putp;

		if (out->q_qinfo->qi_putp == fr_qout) {
			for (qf2 = qif_head; qf2; qf2 = qf2->qf_next)
				if (qf2->qf_wqinfo == out->q_qinfo) {
					qif->qf_outp = qf2->qf_outp;
					break;
				}
			if (!qf2) {
#ifdef	IPFDEBUG
				cmn_err(CE_WARN, "wq:%s put %x qi %x",
					il->ill_name, out->q_qinfo->qi_putp,
					out->q_qinfo);
#endif
				mutex_exit(&ipfs_mutex);
				KFREE(qif);
				continue;
			}
		} else
			qif->qf_outp = out->q_qinfo->qi_putp;

		qif->qf_ill = il;
		qif->qf_iptr = in->q_ptr;
		qif->qf_optr = out->q_ptr;
		qif->qf_hl = il->ill_hdr_length;
		strncpy(qif->qf_name, il->ill_name, sizeof(qif->qf_name));
		qif->qf_name[sizeof(qif->qf_name) - 1] = '\0';
		qif->qf_wqinfo = out->q_qinfo;
		qif->qf_rqinfo = in->q_qinfo;

		qif->qf_next = qif_head;
		qif_head = qif;
		in->q_qinfo->qi_putp = fr_qin;
		out->q_qinfo->qi_putp = fr_qout;
		mutex_exit(&ipfs_mutex);
		cmn_err(CE_CONT, "IP Filter attached to interface [%s]\n",
			qif->qf_name);
	}
	if (!qif_head) {
		cmn_err(CE_CONT, "ipf not attached to any interfaces\n");
		return -1;
	}
	return 0;
}


/*
 * look for bad consistancies between the list of interfaces the filter knows
 * about and those which are currently configured.
 */
int ipfsync()
{
	register struct frentry *f, **fp;
	register qif_t *qif, **qp;
	register ill_t *il;

	mutex_enter(&ipfs_mutex);
	for (qp = &qif_head; (qif = *qp); ) {
		for (il = ill_g_head; il; il = il->ill_next)
			if (qif->qf_ill == il)
				break;
		if (il) {
			qp = &qif->qf_next;
			qif = *qp;
			continue;
		}
		cmn_err(CE_CONT, "detaching [%s]\n", qif->qf_name);
		*qp = qif->qf_next;

		/*
		 * Delete any rules directly associated with this interface
		 */
		mutex_enter(&ipf_mutex);
		for (fp = &ipfilter[0][0]; (f = *fp); )
			if ((void *)f->fr_ifa == (void *)qif->qf_ill) {
				*fp = f->fr_next;
				KFREE(f);
			} else
				fp = &f->fr_next;
		for (fp = &ipfilter[1][0]; (f = *fp); )
			if ((void *)f->fr_ifa == (void *)qif->qf_ill) {
				*fp = f->fr_next;
				KFREE(f);
			} else
				fp = &f->fr_next;
		mutex_exit(&ipf_mutex);

		KFREE(qif);
		qif = *qp;
	}
	mutex_exit(&ipfs_mutex);
	return solattach();
}


/*
 * unhook the IP filter from all defined interfaces with IP addresses
 */
int soldetach()
{
	queue_t *in, *out;
	qif_t *qif, **nq, *qf2;
	ill_t *il;

	for (nq = &qif_head; (qif = *nq); ) {
		for (il = ill_g_head; il; il = il->ill_next)
			if (qif->qf_ill == il)
				break;
		if (il) {
			in = il->ill_rq;
			out = il->ill_wq->q_next;

			mutex_enter(&ipfs_mutex);

			*nq = qif->qf_next;
			in->q_qinfo->qi_putp = qif->qf_inp;
			/*
			 * and change back if something is still interested
			 * in filtering (read side) on this interface.
			 */
			for (qf2 = qif_head; qf2; qf2 = qf2->qf_next)
				if ((qf2 != qif) &&
				    (qf2->qf_rqinfo == in->q_qinfo)) {
					in->q_qinfo->qi_putp = fr_qin;
					break;
				}
			/*
			 * and the write queue...
			 */
			out->q_qinfo->qi_putp = qif->qf_outp;
			for (qf2 = qif_head; qf2; qf2 = qf2->qf_next)
				if ((qf2 != qif) &&
				    (qf2->qf_wqinfo == out->q_qinfo)) {
					out->q_qinfo->qi_putp = fr_qout;
					break;
				}

			KFREE(qif);
			mutex_exit(&ipfs_mutex);
		}
	}
	return ipldetach();
}
